@@ -52,6 +52,16 @@ describe "The rescue keyword" do
       RescueSpecs::SafeNavigationSetterCaptor.should_capture_exception
     end
 
+    it 'using a safely navigated setter method on a nil target' do
+      target = nil
+      begin
+        raise SpecificExampleException, "Raising this to be handled below"
+      rescue SpecificExampleException => target&.captured_error
+        :caught
+      end.should == :caught
+      target.should be_nil
+    end
+
     it 'using a setter method' do
       RescueSpecs::SetterCaptor.should_capture_exception
     end
@@ -563,10 +573,8 @@ describe "The rescue keyword" do
         end
       end
       line = __LINE__
-      foo.should == [
-        "#{__FILE__}:#{line-3}:in 'foo'",
-        "#{__FILE__}:#{line+1}:in 'block (3 levels) in <top (required)>'"
-      ]
+      foo[0].should =~ /#{__FILE__}:#{line-3}:in 'foo'/
+      foo[1].should =~ /#{__FILE__}:#{line+2}:in 'block/
     end
   end
 
