@@ -2,7 +2,9 @@
 require 'test/unit'
 
 class TestEnv < Test::Unit::TestCase
-  IGNORE_CASE = /bccwin|mswin|mingw/ =~ RUBY_PLATFORM
+  windows = /bccwin|mswin|mingw/ =~ RUBY_PLATFORM
+  IGNORE_CASE = windows
+  ENCODING = windows ? Encoding::UTF_8 : Encoding.find("locale")
   PATH_ENV = "PATH"
   INVALID_ENVVARS = [
     "foo\0bar",
@@ -345,12 +347,23 @@ class TestEnv < Test::Unit::TestCase
     ENV["foo"] = "bar"
     ENV["baz"] = "qux"
     s = ENV.inspect
+    expected = [%("foo" => "bar"), %("baz" => "qux")]
+    unless s.start_with?(/\{"foo"/i)
+      expected.reverse!
+    end
+    expected = '{' + expected.join(', ') + '}'
     if IGNORE_CASE
       s = s.upcase
-      assert(s == '{"FOO"=>"BAR", "BAZ"=>"QUX"}' || s == '{"BAZ"=>"QUX", "FOO"=>"BAR"}')
-    else
-      assert(s == '{"foo"=>"bar", "baz"=>"qux"}' || s == '{"baz"=>"qux", "foo"=>"bar"}')
+      expected = expected.upcase
     end
+    assert_equal(expected, s)
+  end
+
+  def test_inspect_encoding
+    ENV.clear
+    key = "VAR\u{e5 e1 e2 e4 e3 101 3042}"
+    ENV[key] = "foo"
+    assert_equal(%{{#{(key.encode(ENCODING) rescue key.b).inspect} => "foo"}}, ENV.inspect)
   end
 
   def test_to_a
@@ -359,12 +372,7 @@ class TestEnv < Test::Unit::TestCase
     ENV["baz"] = "qux"
     a = ENV.to_a
     assert_equal(2, a.size)
-    if IGNORE_CASE
-      a = a.map {|x| x.map {|y| y.upcase } }
-      assert(a == [%w(FOO BAR), %w(BAZ QUX)] || a == [%w(BAZ QUX), %w(FOO BAR)])
-    else
-      assert(a == [%w(foo bar), %w(baz qux)] || a == [%w(baz qux), %w(foo bar)])
-    end
+    check([%w(baz qux), %w(foo bar)], a)
   end
 
   def test_rehash
@@ -403,8 +411,7 @@ class TestEnv < Test::Unit::TestCase
       assert_equal("foo", v)
     end
     assert_invalid_env {|var| ENV.assoc(var)}
-    encoding = /mswin|mingw/ =~ RUBY_PLATFORM ? Encoding::UTF_8 : Encoding.find("locale")
-    assert_equal(encoding, v.encoding)
+    assert_equal(ENCODING, v.encoding)
   end
 
   def test_has_value2
@@ -450,13 +457,14 @@ class TestEnv < Test::Unit::TestCase
     assert_equal(h1, h2)
   end
 
-  def check(as, bs)
+  def assert_equal_env(as, bs)
     if IGNORE_CASE
       as = as.map {|k, v| [k.upcase, v] }
       bs = bs.map {|k, v| [k.upcase, v] }
     end
     assert_equal(as.sort, bs.sort)
   end
+  alias check assert_equal_env
 
   def test_shift
     ENV.clear
@@ -517,7 +525,7 @@ class TestEnv < Test::Unit::TestCase
     assert_equal(huge_value, ENV["foo"])
   end
 
-  if /mswin|mingw/ =~ RUBY_PLATFORM
+  if windows
     def windows_version
       @windows_version ||= %x[ver][/Version (\d+)/, 1].to_i
     end
@@ -1089,12 +1097,16 @@ class TestEnv < Test::Unit::TestCase
         Ractor.yield s
       end
       s = r.take
+      expected = ['"foo" => "bar"', '"baz" => "qux"']
+      unless s.start_with?(/\{"foo"/i)
+        expected.reverse!
+      end
+      expected = "{" + expected.join(', ') + "}"
       if #{ignore_case_str}
         s = s.upcase
-        assert(s == '{"FOO"=>"BAR", "BAZ"=>"QUX"}' || s == '{"BAZ"=>"QUX", "FOO"=>"BAR"}')
-      else
-        assert(s == '{"foo"=>"bar", "baz"=>"qux"}' || s == '{"baz"=>"qux", "foo"=>"bar"}')
+        expected = expected.upcase
       end
+      assert_equal(expected, s)
     end;
   end
 
@@ -1109,12 +1121,13 @@ class TestEnv < Test::Unit::TestCase
       end
       a = r.take
       assert_equal(2, a.size)
+      expected = [%w(baz qux), %w(foo bar)]
       if #{ignore_case_str}
-        a = a.map {|x| x.map {|y| y.upcase } }
-        assert(a == [%w(FOO BAR), %w(BAZ QUX)] || a == [%w(BAZ QUX), %w(FOO BAR)])
-      else
-        assert(a == [%w(foo bar), %w(baz qux)] || a == [%w(baz qux), %w(foo bar)])
+        a = a.map {|x, y| [x.upcase, y]}
+        expected.map! {|x, y| [x.upcase, y]}
       end
+      a.sort!
+      assert_equal(expected, a)
     end;
   end
 
@@ -1479,11 +1492,18 @@ class TestEnv < Test::Unit::TestCase
 
     def test_utf8
       text = "testing \u{e5 e1 e2 e4 e3 101 3042}"
-      test = ENV["test"]
       ENV["test"] = text
       assert_equal text, ENV["test"]
-    ensure
-      ENV["test"] = test
+    end
+
+    def test_utf8_empty
+      key = "VAR\u{e5 e1 e2 e4 e3 101 3042}"
+      ENV[key] = "x"
+      assert_equal "x", ENV[key]
+      ENV[key] = ""
+      assert_equal "", ENV[key]
+      ENV[key] = nil
+      assert_nil ENV[key]
     end
   end
 end
