@@ -2,11 +2,11 @@ use std::collections::HashMap;
 use std::fmt;
 use std::convert::From;
 use std::mem::take;
-use crate::codegen::{gen_outlined_exit, gen_counted_exit};
-use crate::cruby::{vm_stack_canary, SIZEOF_VALUE_I32, VALUE};
+use crate::codegen::{gen_counted_exit, gen_outlined_exit};
+use crate::cruby::{vm_stack_canary, SIZEOF_VALUE_I32, VALUE, VM_ENV_DATA_SIZE};
 use crate::virtualmem::CodePtr;
 use crate::asm::{CodeBlock, OutlinedCb};
-use crate::core::{Context, RegTemps, MAX_REG_TEMPS};
+use crate::core::{Context, RegMapping, RegOpnd, MAX_CTX_TEMPS};
 use crate::options::*;
 use crate::stats::*;
 
@@ -77,10 +77,12 @@ pub enum Opnd
         num_bits: u8,
         /// ctx.stack_size when this operand is made. Used with idx for Opnd::Reg.
         stack_size: u8,
+        /// The number of local variables in the current ISEQ. Used only for locals.
+        num_locals: Option<u32>,
         /// ctx.sp_offset when this operand is made. Used with idx for Opnd::Mem.
         sp_offset: i8,
-        /// ctx.reg_temps when this operand is read. Used for register allocation.
-        reg_temps: Option<RegTemps>
+        /// ctx.reg_mapping when this operand is read. Used for register allocation.
+        reg_mapping: Option<RegMapping>
     },
 
     // Low-level operands, for lowering
@@ -172,7 +174,7 @@ impl Opnd
             Opnd::Reg(reg) => Some(Opnd::Reg(reg.with_num_bits(num_bits))),
             Opnd::Mem(Mem { base, disp, .. }) => Some(Opnd::Mem(Mem { base, disp, num_bits })),
             Opnd::InsnOut { idx, .. } => Some(Opnd::InsnOut { idx, num_bits }),
-            Opnd::Stack { idx, stack_size, sp_offset, reg_temps, .. } => Some(Opnd::Stack { idx, num_bits, stack_size, sp_offset, reg_temps }),
+            Opnd::Stack { idx, stack_size, num_locals, sp_offset, reg_mapping, .. } => Some(Opnd::Stack { idx, num_bits, stack_size, num_locals, sp_offset, reg_mapping }),
             _ => None,
         }
     }
@@ -227,28 +229,26 @@ impl Opnd
         Self::match_num_bits_iter(opnds.iter())
     }
 
-    /// Calculate Opnd::Stack's index from the stack bottom.
-    pub fn stack_idx(&self) -> u8 {
-        self.get_stack_idx().unwrap()
+    /// Convert Opnd::Stack into RegMapping
+    pub fn reg_opnd(&self) -> RegOpnd {
+        self.get_reg_opnd().unwrap()
     }
 
-    /// Calculate Opnd::Stack's index from the stack bottom if it's Opnd::Stack.
-    pub fn get_stack_idx(&self) -> Option<u8> {
-        match self {
-            Opnd::Stack { idx, stack_size, .. } => {
-                Some((*stack_size as isize - *idx as isize - 1) as u8)
-            },
-            _ => None
-        }
-    }
-
-    /// Get the index for stack temp registers.
-    pub fn reg_idx(&self) -> usize {
-        match self {
-            Opnd::Stack { .. } => {
-                self.stack_idx() as usize % get_option!(num_temp_regs)
-            },
-            _ => unreachable!(),
+    /// Convert an operand into RegMapping if it's Opnd::Stack
+    pub fn get_reg_opnd(&self) -> Option<RegOpnd> {
+        match *self {
+            Opnd::Stack { idx, stack_size, num_locals, .. } => Some(
+                if let Some(num_locals) = num_locals {
+                    let last_idx = stack_size as i32 + VM_ENV_DATA_SIZE as i32 - 1;
+                    assert!(last_idx <= idx, "Local index {} must be >= last local index {}", idx, last_idx);
+                    assert!(idx <= last_idx + num_locals as i32, "Local index {} must be < last local index {} + local size {}", idx, last_idx, num_locals);
+                    RegOpnd::Local((last_idx + num_locals as i32 - idx) as u8)
+                } else {
+                    assert!(idx < stack_size as i32);
+                    RegOpnd::Stack((stack_size as i32 - idx - 1) as u8)
+                }
+            ),
+            _ => None,
         }
     }
 }
@@ -974,7 +974,7 @@ pub struct SideExitContext {
     /// Context fields used by get_generic_ctx()
     pub stack_size: u8,
     pub sp_offset: i8,
-    pub reg_temps: RegTemps,
+    pub reg_mapping: RegMapping,
     pub is_return_landing: bool,
     pub is_deferred: bool,
 }
@@ -986,7 +986,7 @@ impl SideExitContext {
             pc,
             stack_size: ctx.get_stack_size(),
             sp_offset: ctx.get_sp_offset(),
-            reg_temps: ctx.get_reg_temps(),
+            reg_mapping: ctx.get_reg_mapping(),
             is_return_landing: ctx.is_return_landing(),
             is_deferred: ctx.is_deferred(),
         };
@@ -1002,7 +1002,7 @@ impl SideExitContext {
         let mut ctx = Context::default();
         ctx.set_stack_size(self.stack_size);
         ctx.set_sp_offset(self.sp_offset);
-        ctx.set_reg_temps(self.reg_temps);
+        ctx.set_reg_mapping(self.reg_mapping);
         if self.is_return_landing {
             ctx.set_as_return_landing();
         }
@@ -1031,6 +1031,13 @@ pub struct Assembler {
     /// Context for generating the current insn
     pub ctx: Context,
 
+    /// The current ISEQ's local table size. asm.local_opnd() uses this, and it's
+    /// sometimes hard to pass this value, e.g. asm.spill_regs() in asm.ccall().
+    ///
+    /// `None` means we're not assembling for an ISEQ, or that the local size is
+    /// not relevant.
+    pub(super) num_locals: Option<u32>,
+
     /// Side exit caches for each SideExitContext
     pub(super) side_exits: HashMap<SideExitContext, CodePtr>,
 
@@ -1046,16 +1053,31 @@ pub struct Assembler {
 
 impl Assembler
 {
-    pub fn new() -> Self {
-        Self::new_with_label_names(Vec::default(), HashMap::default())
+    /// Create an Assembler for ISEQ-specific code.
+    /// It includes all inline code and some outlined code like side exits and stubs.
+    pub fn new(num_locals: u32) -> Self {
+        Self::new_with_label_names(Vec::default(), HashMap::default(), Some(num_locals))
     }
 
-    pub fn new_with_label_names(label_names: Vec<String>, side_exits: HashMap<SideExitContext, CodePtr>) -> Self {
+    /// Create an Assembler for outlined code that are not specific to any ISEQ,
+    /// e.g. trampolines that are shared globally.
+    pub fn new_without_iseq() -> Self {
+        Self::new_with_label_names(Vec::default(), HashMap::default(), None)
+    }
+
+    /// Create an Assembler with parameters that are populated by another Assembler instance.
+    /// This API is used for copying an Assembler for the next compiler pass.
+    pub fn new_with_label_names(
+        label_names: Vec<String>,
+        side_exits: HashMap<SideExitContext, CodePtr>,
+        num_locals: Option<u32>
+    ) -> Self {
         Self {
             insns: Vec::with_capacity(ASSEMBLER_INSNS_CAPACITY),
             live_ranges: Vec::with_capacity(ASSEMBLER_INSNS_CAPACITY),
             label_names,
             ctx: Context::default(),
+            num_locals,
             side_exits,
             side_exit_pc: None,
             side_exit_stack_size: None,
@@ -1069,6 +1091,11 @@ impl Assembler
         &TEMP_REGS[0..num_regs]
     }
 
+    /// Get the number of locals for the ISEQ being compiled
+    pub fn get_num_locals(&self) -> Option<u32> {
+        self.num_locals
+    }
+
     /// Set a context for generating side exits
     pub fn set_side_exit_context(&mut self, pc: *mut VALUE, stack_size: u8) {
         self.side_exit_pc = Some(pc);
@@ -1090,31 +1117,32 @@ impl Assembler
 
         let mut opnd_iter = insn.opnd_iter_mut();
         while let Some(opnd) = opnd_iter.next() {
-            match opnd {
+            match *opnd {
                 // If we find any InsnOut from previous instructions, we're going to update
                 // the live range of the previous instruction to point to this one.
                 Opnd::InsnOut { idx, .. } => {
-                    assert!(*idx < self.insns.len());
-                    self.live_ranges[*idx] = insn_idx;
+                    assert!(idx < self.insns.len());
+                    self.live_ranges[idx] = insn_idx;
                 }
                 Opnd::Mem(Mem { base: MemBase::InsnOut(idx), .. }) => {
-                    assert!(*idx < self.insns.len());
-                    self.live_ranges[*idx] = insn_idx;
+                    assert!(idx < self.insns.len());
+                    self.live_ranges[idx] = insn_idx;
                 }
-                // Set current ctx.reg_temps to Opnd::Stack.
-                Opnd::Stack { idx, num_bits, stack_size, sp_offset, reg_temps: None } => {
+                // Set current ctx.reg_mapping to Opnd::Stack.
+                Opnd::Stack { idx, num_bits, stack_size, num_locals, sp_offset, reg_mapping: None } => {
                     assert_eq!(
                         self.ctx.get_stack_size() as i16 - self.ctx.get_sp_offset() as i16,
-                        *stack_size as i16 - *sp_offset as i16,
+                        stack_size as i16 - sp_offset as i16,
                         "Opnd::Stack (stack_size: {}, sp_offset: {}) expects a different SP position from asm.ctx (stack_size: {}, sp_offset: {})",
-                        *stack_size, *sp_offset, self.ctx.get_stack_size(), self.ctx.get_sp_offset(),
+                        stack_size, sp_offset, self.ctx.get_stack_size(), self.ctx.get_sp_offset(),
                     );
                     *opnd = Opnd::Stack {
-                        idx: *idx,
-                        num_bits: *num_bits,
-                        stack_size: *stack_size,
-                        sp_offset: *sp_offset,
-                        reg_temps: Some(self.ctx.get_reg_temps()),
+                        idx,
+                        num_bits,
+                        stack_size,
+                        num_locals,
+                        sp_offset,
+                        reg_mapping: Some(self.ctx.get_reg_mapping()),
                     };
                 }
                 _ => {}
@@ -1141,7 +1169,7 @@ impl Assembler
         // Get a cached side exit
         let side_exit = match self.side_exits.get(&side_exit_context) {
             None => {
-                let exit_code = gen_outlined_exit(side_exit_context.pc, &side_exit_context.get_ctx(), ocb)?;
+                let exit_code = gen_outlined_exit(side_exit_context.pc, self.num_locals.unwrap(), &side_exit_context.get_ctx(), ocb)?;
                 self.side_exits.insert(*side_exit_context, exit_code);
                 exit_code
             }
@@ -1175,20 +1203,20 @@ impl Assembler
         }
 
         // Convert Opnd::Stack to Opnd::Reg
-        fn reg_opnd(opnd: &Opnd) -> Opnd {
+        fn reg_opnd(opnd: &Opnd, reg_idx: usize) -> Opnd {
             let regs = Assembler::get_temp_regs();
             if let Opnd::Stack { num_bits, .. } = *opnd {
                 incr_counter!(temp_reg_opnd);
-                Opnd::Reg(regs[opnd.reg_idx()]).with_num_bits(num_bits).unwrap()
+                Opnd::Reg(regs[reg_idx]).with_num_bits(num_bits).unwrap()
             } else {
                 unreachable!()
             }
         }
 
         match opnd {
-            Opnd::Stack { reg_temps, .. } => {
-                if opnd.stack_idx() < MAX_REG_TEMPS && reg_temps.unwrap().get(opnd.stack_idx()) {
-                    reg_opnd(opnd)
+            Opnd::Stack { reg_mapping, .. } => {
+                if let Some(reg_idx) = reg_mapping.unwrap().get_reg(opnd.reg_opnd()) {
+                    reg_opnd(opnd, reg_idx)
                 } else {
                     mem_opnd(opnd)
                 }
@@ -1198,18 +1226,11 @@ impl Assembler
     }
 
     /// Allocate a register to a stack temp if available.
-    pub fn alloc_temp_reg(&mut self, stack_idx: u8) {
-        if get_option!(num_temp_regs) == 0 {
-            return;
-        }
-
+    pub fn alloc_reg(&mut self, mapping: RegOpnd) {
         // Allocate a register if there's no conflict.
-        let mut reg_temps = self.ctx.get_reg_temps();
-        if reg_temps.conflicts_with(stack_idx) {
-            assert!(!reg_temps.get(stack_idx));
-        } else {
-            reg_temps.set(stack_idx, true);
-            self.set_reg_temps(reg_temps);
+        let mut reg_mapping = self.ctx.get_reg_mapping();
+        if reg_mapping.alloc_reg(mapping) {
+            self.set_reg_mapping(reg_mapping);
         }
     }
 
@@ -1220,47 +1241,96 @@ impl Assembler
         self.ctx.clear_local_types();
     }
 
-    /// Spill all live stack temps from registers to the stack
-    pub fn spill_temps(&mut self) {
-        // Forget registers above the stack top
-        let mut reg_temps = self.ctx.get_reg_temps();
-        for stack_idx in self.ctx.get_stack_size()..MAX_REG_TEMPS {
-            reg_temps.set(stack_idx, false);
-        }
-        self.set_reg_temps(reg_temps);
+    /// Repurpose stack temp registers to the corresponding locals for arguments
+    pub fn map_temp_regs_to_args(&mut self, callee_ctx: &mut Context, argc: i32) -> Vec<RegOpnd> {
+        let mut callee_reg_mapping = callee_ctx.get_reg_mapping();
+        let mut mapped_temps = vec![];
 
-        // Spill live stack temps
-        if self.ctx.get_reg_temps() != RegTemps::default() {
-            asm_comment!(self, "spill_temps: {:08b} -> {:08b}", self.ctx.get_reg_temps().as_u8(), RegTemps::default().as_u8());
-            for stack_idx in 0..u8::min(MAX_REG_TEMPS, self.ctx.get_stack_size()) {
-                if self.ctx.get_reg_temps().get(stack_idx) {
-                    let idx = self.ctx.get_stack_size() - 1 - stack_idx;
-                    self.spill_temp(self.stack_opnd(idx.into()));
-                    reg_temps.set(stack_idx, false);
-                }
+        for arg_idx in 0..argc {
+            let stack_idx: u8 = (self.ctx.get_stack_size() as i32 - argc + arg_idx).try_into().unwrap();
+            let temp_opnd = RegOpnd::Stack(stack_idx);
+
+            // For each argument, if the stack temp for it has a register,
+            // let the callee use the register for the local variable.
+            if let Some(reg_idx) = self.ctx.get_reg_mapping().get_reg(temp_opnd) {
+                let local_opnd = RegOpnd::Local(arg_idx.try_into().unwrap());
+                callee_reg_mapping.set_reg(local_opnd, reg_idx);
+                mapped_temps.push(temp_opnd);
             }
-            self.ctx.set_reg_temps(reg_temps);
         }
 
-        // Every stack temp should have been spilled
-        assert_eq!(self.ctx.get_reg_temps(), RegTemps::default());
+        asm_comment!(self, "local maps: {:?}", callee_reg_mapping);
+        callee_ctx.set_reg_mapping(callee_reg_mapping);
+        mapped_temps
+    }
+
+    /// Spill all live registers to the stack
+    pub fn spill_regs(&mut self) {
+        self.spill_regs_except(&vec![]);
+    }
+
+    /// Spill all live registers except `ignored_temps` to the stack
+    pub fn spill_regs_except(&mut self, ignored_temps: &Vec<RegOpnd>) {
+        // Forget registers above the stack top
+        let mut reg_mapping = self.ctx.get_reg_mapping();
+        for stack_idx in self.ctx.get_stack_size()..MAX_CTX_TEMPS as u8 {
+            reg_mapping.dealloc_reg(RegOpnd::Stack(stack_idx));
+        }
+        self.set_reg_mapping(reg_mapping);
+
+        // If no registers are in use, skip all checks
+        if self.ctx.get_reg_mapping() == RegMapping::default() {
+            return;
+        }
+
+        // Collect stack temps to be spilled
+        let mut spilled_opnds = vec![];
+        for stack_idx in 0..u8::min(MAX_CTX_TEMPS as u8, self.ctx.get_stack_size()) {
+            let reg_opnd = RegOpnd::Stack(stack_idx);
+            if !ignored_temps.contains(&reg_opnd) && reg_mapping.dealloc_reg(reg_opnd) {
+                let idx = self.ctx.get_stack_size() - 1 - stack_idx;
+                let spilled_opnd = self.stack_opnd(idx.into());
+                spilled_opnds.push(spilled_opnd);
+                reg_mapping.dealloc_reg(spilled_opnd.reg_opnd());
+            }
+        }
+
+        // Collect locals to be spilled
+        for local_idx in 0..MAX_CTX_TEMPS as u8 {
+            if reg_mapping.dealloc_reg(RegOpnd::Local(local_idx)) {
+                let first_local_ep_offset = self.num_locals.unwrap() + VM_ENV_DATA_SIZE - 1;
+                let ep_offset = first_local_ep_offset - local_idx as u32;
+                let spilled_opnd = self.local_opnd(ep_offset);
+                spilled_opnds.push(spilled_opnd);
+                reg_mapping.dealloc_reg(spilled_opnd.reg_opnd());
+            }
+        }
+
+        // Spill stack temps and locals
+        if !spilled_opnds.is_empty() {
+            asm_comment!(self, "spill_regs: {:?} -> {:?}", self.ctx.get_reg_mapping(), reg_mapping);
+            for &spilled_opnd in spilled_opnds.iter() {
+                self.spill_reg(spilled_opnd);
+            }
+            self.ctx.set_reg_mapping(reg_mapping);
+        }
     }
 
     /// Spill a stack temp from a register to the stack
-    fn spill_temp(&mut self, opnd: Opnd) {
-        assert!(self.ctx.get_reg_temps().get(opnd.stack_idx()));
+    pub fn spill_reg(&mut self, opnd: Opnd) {
+        assert_ne!(self.ctx.get_reg_mapping().get_reg(opnd.reg_opnd()), None);
 
-        // Use different RegTemps for dest and src operands
-        let reg_temps = self.ctx.get_reg_temps();
-        let mut mem_temps = reg_temps;
-        mem_temps.set(opnd.stack_idx(), false);
+        // Use different RegMappings for dest and src operands
+        let reg_mapping = self.ctx.get_reg_mapping();
+        let mut mem_mappings = reg_mapping;
+        mem_mappings.dealloc_reg(opnd.reg_opnd());
 
         // Move the stack operand from a register to memory
         match opnd {
-            Opnd::Stack { idx, num_bits, stack_size, sp_offset, .. } => {
+            Opnd::Stack { idx, num_bits, stack_size, num_locals, sp_offset, .. } => {
                 self.mov(
-                    Opnd::Stack { idx, num_bits, stack_size, sp_offset, reg_temps: Some(mem_temps) },
-                    Opnd::Stack { idx, num_bits, stack_size, sp_offset, reg_temps: Some(reg_temps) },
+                    Opnd::Stack { idx, num_bits, stack_size, num_locals, sp_offset, reg_mapping: Some(mem_mappings) },
+                    Opnd::Stack { idx, num_bits, stack_size, num_locals, sp_offset, reg_mapping: Some(reg_mapping) },
                 );
             }
             _ => unreachable!(),
@@ -1269,21 +1339,47 @@ impl Assembler
     }
 
     /// Update which stack temps are in a register
-    pub fn set_reg_temps(&mut self, reg_temps: RegTemps) {
-        if self.ctx.get_reg_temps() != reg_temps {
-            asm_comment!(self, "reg_temps: {:08b} -> {:08b}", self.ctx.get_reg_temps().as_u8(), reg_temps.as_u8());
-            self.ctx.set_reg_temps(reg_temps);
-            self.verify_reg_temps();
+    pub fn set_reg_mapping(&mut self, reg_mapping: RegMapping) {
+        if self.ctx.get_reg_mapping() != reg_mapping {
+            asm_comment!(self, "reg_mapping: {:?} -> {:?}", self.ctx.get_reg_mapping(), reg_mapping);
+            self.ctx.set_reg_mapping(reg_mapping);
         }
     }
 
-    /// Assert there's no conflict in stack temp register allocation
-    fn verify_reg_temps(&self) {
-        for stack_idx in 0..MAX_REG_TEMPS {
-            if self.ctx.get_reg_temps().get(stack_idx) {
-                assert!(!self.ctx.get_reg_temps().conflicts_with(stack_idx));
+    // Shuffle register moves, sometimes adding extra moves using SCRATCH_REG,
+    // so that they will not rewrite each other before they are used.
+    pub fn reorder_reg_moves(old_moves: &Vec<(Reg, Opnd)>) -> Vec<(Reg, Opnd)> {
+        // Return the index of a move whose destination is not used as a source if any.
+        fn find_safe_move(moves: &Vec<(Reg, Opnd)>) -> Option<usize> {
+            moves.iter().enumerate().find(|(_, &(dest_reg, _))| {
+                moves.iter().all(|&(_, src_opnd)| src_opnd != Opnd::Reg(dest_reg))
+            }).map(|(index, _)| index)
+        }
+
+        // Remove moves whose source and destination are the same
+        let mut old_moves: Vec<(Reg, Opnd)> = old_moves.clone().into_iter()
+            .filter(|&(reg, opnd)| Opnd::Reg(reg) != opnd).collect();
+
+        let mut new_moves = vec![];
+        while old_moves.len() > 0 {
+            // Keep taking safe moves
+            while let Some(index) = find_safe_move(&old_moves) {
+                new_moves.push(old_moves.remove(index));
+            }
+
+            // No safe move. Load the source of one move into SCRATCH_REG, and
+            // then load SCRATCH_REG into the destination when it's safe.
+            if old_moves.len() > 0 {
+                // Make sure it's safe to use SCRATCH_REG
+                assert!(old_moves.iter().all(|&(_, opnd)| opnd != Opnd::Reg(Assembler::SCRATCH_REG)));
+
+                // Move SCRATCH <- opnd, and delay reg <- SCRATCH
+                let (reg, opnd) = old_moves.remove(0);
+                new_moves.push((Assembler::SCRATCH_REG, opnd));
+                old_moves.push((reg, Opnd::Reg(Assembler::SCRATCH_REG)));
             }
         }
+        new_moves
     }
 
     /// Sets the out field on the various instructions that require allocated
@@ -1331,42 +1427,6 @@ impl Assembler
             }
         }
 
-        // Reorder C argument moves, sometimes adding extra moves using SCRATCH_REG,
-        // so that they will not rewrite each other before they are used.
-        fn reorder_c_args(c_args: &Vec<(Reg, Opnd)>) -> Vec<(Reg, Opnd)> {
-            // Return the index of a move whose destination is not used as a source if any.
-            fn find_safe_arg(c_args: &Vec<(Reg, Opnd)>) -> Option<usize> {
-                c_args.iter().enumerate().find(|(_, &(dest_reg, _))| {
-                    c_args.iter().all(|&(_, src_opnd)| src_opnd != Opnd::Reg(dest_reg))
-                }).map(|(index, _)| index)
-            }
-
-            // Remove moves whose source and destination are the same
-            let mut c_args: Vec<(Reg, Opnd)> = c_args.clone().into_iter()
-                .filter(|&(reg, opnd)| Opnd::Reg(reg) != opnd).collect();
-
-            let mut moves = vec![];
-            while c_args.len() > 0 {
-                // Keep taking safe moves
-                while let Some(index) = find_safe_arg(&c_args) {
-                    moves.push(c_args.remove(index));
-                }
-
-                // No safe move. Load the source of one move into SCRATCH_REG, and
-                // then load SCRATCH_REG into the destination when it's safe.
-                if c_args.len() > 0 {
-                    // Make sure it's safe to use SCRATCH_REG
-                    assert!(c_args.iter().all(|&(_, opnd)| opnd != Opnd::Reg(Assembler::SCRATCH_REG)));
-
-                    // Move SCRATCH <- opnd, and delay reg <- SCRATCH
-                    let (reg, opnd) = c_args.remove(0);
-                    moves.push((Assembler::SCRATCH_REG, opnd));
-                    c_args.push((reg, Opnd::Reg(Assembler::SCRATCH_REG)));
-                }
-            }
-            moves
-        }
-
         // Adjust the number of entries in live_ranges so that it can be indexed by mapped indexes.
         fn shift_live_ranges(live_ranges: &mut Vec<usize>, start_index: usize, shift_offset: isize) {
             if shift_offset >= 0 {
@@ -1411,7 +1471,7 @@ impl Assembler
         let live_ranges: Vec<usize> = take(&mut self.live_ranges);
         // shifted_live_ranges is indexed by mapped indexes in insn operands.
         let mut shifted_live_ranges: Vec<usize> = live_ranges.clone();
-        let mut asm = Assembler::new_with_label_names(take(&mut self.label_names), take(&mut self.side_exits));
+        let mut asm = Assembler::new_with_label_names(take(&mut self.label_names), take(&mut self.side_exits), self.num_locals);
         let mut iterator = self.into_draining_iter();
 
         while let Some((index, mut insn)) = iterator.next_mapped() {
@@ -1542,7 +1602,7 @@ impl Assembler
                 if c_args.len() > 0 {
                     // Resolve C argument dependencies
                     let c_args_len = c_args.len() as isize;
-                    let moves = reorder_c_args(&c_args.drain(..).into_iter().collect());
+                    let moves = Self::reorder_reg_moves(&c_args.drain(..).into_iter().collect());
                     shift_live_ranges(&mut shifted_live_ranges, asm.insns.len(), moves.len() as isize - c_args_len);
 
                     // Push batched C arguments
@@ -1565,13 +1625,10 @@ impl Assembler
     #[must_use]
     pub fn compile(self, cb: &mut CodeBlock, ocb: Option<&mut OutlinedCb>) -> Option<(CodePtr, Vec<u32>)>
     {
-        #[cfg(feature = "disasm")]
         let start_addr = cb.get_write_ptr();
-
         let alloc_regs = Self::get_alloc_regs();
         let ret = self.compile_with_regs(cb, ocb, alloc_regs);
 
-        #[cfg(feature = "disasm")]
         if let Some(dump_disasm) = get_option_ref!(dump_disasm) {
             use crate::disasm::dump_disasm_addr_range;
             let end_addr = cb.get_write_ptr();
@@ -1706,24 +1763,24 @@ impl Assembler {
         // Let vm_check_canary() assert this ccall's leafness if leaf_ccall is set
         let canary_opnd = self.set_stack_canary(&opnds);
 
-        let old_temps = self.ctx.get_reg_temps(); // with registers
+        let old_temps = self.ctx.get_reg_mapping(); // with registers
         // Spill stack temp registers since they are caller-saved registers.
         // Note that this doesn't spill stack temps that are already popped
         // but may still be used in the C arguments.
-        self.spill_temps();
-        let new_temps = self.ctx.get_reg_temps(); // all spilled
+        self.spill_regs();
+        let new_temps = self.ctx.get_reg_mapping(); // all spilled
 
-        // Temporarily manipulate RegTemps so that we can use registers
+        // Temporarily manipulate RegMappings so that we can use registers
         // to pass stack operands that are already spilled above.
-        self.ctx.set_reg_temps(old_temps);
+        self.ctx.set_reg_mapping(old_temps);
 
         // Call a C function
         let out = self.next_opnd_out(Opnd::match_num_bits(&opnds));
         self.push_insn(Insn::CCall { fptr, opnds, out });
 
         // Registers in old_temps may be clobbered by the above C call,
-        // so rollback the manipulated RegTemps to a spilled version.
-        self.ctx.set_reg_temps(new_temps);
+        // so rollback the manipulated RegMappings to a spilled version.
+        self.ctx.set_reg_mapping(new_temps);
 
         // Clear the canary after use
         if let Some(canary_opnd) = canary_opnd {
@@ -1740,8 +1797,8 @@ impl Assembler {
 
         // If the slot is already used, which is a valid optimization to avoid spills,
         // give up the verification.
-        let canary_opnd = if cfg!(debug_assertions) && self.leaf_ccall && opnds.iter().all(|opnd|
-            opnd.get_stack_idx() != canary_opnd.get_stack_idx()
+        let canary_opnd = if cfg!(feature = "runtime_checks") && self.leaf_ccall && opnds.iter().all(|opnd|
+            opnd.get_reg_opnd() != canary_opnd.get_reg_opnd()
         ) {
             asm_comment!(self, "set stack canary");
             self.mov(canary_opnd, vm_stack_canary().into());
@@ -1770,9 +1827,9 @@ impl Assembler {
     pub fn cpop_all(&mut self) {
         self.push_insn(Insn::CPopAll);
 
-        // Re-enable ccall's RegTemps assertion disabled by cpush_all.
+        // Re-enable ccall's RegMappings assertion disabled by cpush_all.
         // cpush_all + cpop_all preserve all stack temp registers, so it's safe.
-        self.set_reg_temps(self.ctx.get_reg_temps());
+        self.set_reg_mapping(self.ctx.get_reg_mapping());
     }
 
     pub fn cpop_into(&mut self, opnd: Opnd) {
@@ -1789,8 +1846,8 @@ impl Assembler {
         // Mark all temps as not being in registers.
         // Temps will be marked back as being in registers by cpop_all.
         // We assume that cpush_all + cpop_all are used for C functions in utils.rs
-        // that don't require spill_temps for GC.
-        self.set_reg_temps(RegTemps::default());
+        // that don't require spill_regs for GC.
+        self.set_reg_mapping(RegMapping::default());
     }
 
     pub fn cret(&mut self, opnd: Opnd) {
@@ -2057,10 +2114,10 @@ impl Assembler {
 }
 
 /// Macro to use format! for Insn::Comment, which skips a format! call
-/// when disasm is not supported.
+/// when not dumping disassembly.
 macro_rules! asm_comment {
     ($asm:expr, $($fmt:tt)*) => {
-        if cfg!(feature = "disasm") {
+        if $crate::options::get_option_ref!(dump_disasm).is_some() {
             $asm.push_insn(Insn::Comment(format!($($fmt)*)));
         }
     };
